﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using PaintDotNet;

namespace BeeDevelopment.ZSoftPcxPlugin {

	public class ZSoftPcxFileType : FileType {

		public ZSoftPcxFileType()
			: base("ZSoft PCX", FileTypeFlags.SupportsLoading, new string[] { ".pcx" }) {
		}

		protected override Document OnLoad(Stream input) {

			BinaryReader pcxReader = new BinaryReader(input);

			#region Header

			// Check manufacturer code.
			if (pcxReader.ReadByte() != 10) {
				MessageBox.Show("Invalid manufacturer flag.", "ZSoft PCX", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return null;
			}

			// Version.
			var version = pcxReader.ReadByte();

			// Get compression scheme.
			byte compression;
			if ((compression = pcxReader.ReadByte()) > 1) {
				MessageBox.Show(string.Format("Encoding type {0} not supported."), "ZSoft PCX", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return null;
			}

			// Bits-per-pixel-per-plane.
			byte bitsPerPixel;
			if ((bitsPerPixel = pcxReader.ReadByte()) != 8) {
				MessageBox.Show(string.Format("Only 8 bits per pixel images are supported (file specifies {0}).", bitsPerPixel), "ZSoft PCX", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return null;
			}

			// Skip window.
			short xMin = pcxReader.ReadInt16();
			short yMin = pcxReader.ReadInt16();
			short xMax = pcxReader.ReadInt16();
			short yMax = pcxReader.ReadInt16();

			int width = (xMax - xMin) + 1;
			int height = (yMax - yMin) + 1;

			// Resolution of the image.
			ushort horizontalResolution = pcxReader.ReadUInt16();
			ushort verticalResolution = pcxReader.ReadUInt16();

			// Colour map.
			Color[] palette = new Color[16];
			for (int palIndex = 0; palIndex < 16; ++palIndex) palette[palIndex] = Color.FromArgb(255, pcxReader.ReadByte(), pcxReader.ReadByte(), pcxReader.ReadByte());

			// Reserved.
			byte reserved = pcxReader.ReadByte();
			if (reserved != 0) {
				// Quake 2 doesn't zero its PCX struct before writing to disk and leaves reserved uninitialised. Let's not bother with a message.
				// MessageBox.Show(string.Format("Reserved field should be 0 (not {0}).", reserved), "ZSoft PCX", MessageBoxButtons.OK, MessageBoxIcon.Information);
			}

			// Colour planes.
			var pixelFormat = PixelFormat.Undefined;
			byte planeCount;
			switch (planeCount = pcxReader.ReadByte()) {
				case 1:
					pixelFormat = PixelFormat.Format8bppIndexed;
					break;
				case 3:
					pixelFormat = PixelFormat.Format24bppRgb;
					break;
				default:
					MessageBox.Show(string.Format("Only 1 or 3 plane images are supported (file specifies {0}).", planeCount), "ZSoft PCX", MessageBoxButtons.OK, MessageBoxIcon.Error);
					return null;
			}

			// Number of bytes per scanline.
			ushort bytesPerLine = pcxReader.ReadUInt16();

			#endregion

			#region Extended Palette

			if (version == 5 && pixelFormat == PixelFormat.Format8bppIndexed) {
				input.Seek(-768, SeekOrigin.End);
				palette = new Color[256];
				for (int palIndex = 0; palIndex < 256; ++palIndex) palette[palIndex] = Color.FromArgb(255, pcxReader.ReadByte(), pcxReader.ReadByte(), pcxReader.ReadByte());
			}

			#endregion

			#region Decoding Image Data

			// Seek to the data itself:
			input.Seek(128, SeekOrigin.Begin);

			// Create a Bitmap and decode the image into it.
			using (var bitmap = new Bitmap(width, height, pixelFormat)) {

				// Set the palette.
				if (pixelFormat == PixelFormat.Format8bppIndexed) {
					var pal = bitmap.Palette;
					for (int palIndex = 0; palIndex < palette.Length; ++palIndex) {
						pal.Entries[palIndex] = palette[palIndex];
					}
					bitmap.Palette = pal;
				}

				// Lock the bitmap for output.
				var lockedBitmap = bitmap.LockBits(new Rectangle(0, 0, width, height), ImageLockMode.WriteOnly, pixelFormat);
				var outputPtr = lockedBitmap.Scan0;

				// Decompress each scanline in turn.

				try {

					// Storage for decompressed scanline.
					byte[] decompressedScanline = new byte[planeCount * bytesPerLine];
					byte[] rearrangedScanline = planeCount > 1 ? new byte[decompressedScanline.Length] : decompressedScanline;

					// Used to keep track of the RLE compression.
					int runCount = 0;
					byte runValue = 0;

					for (int x = 0, y = 0; y < height; ) {

						// Read a single byte.
						byte value = 0;
						switch (compression) {
							case 0:
								value = pcxReader.ReadByte();
								break;
							case 1:
								if (--runCount > 0) {
									value = runValue;
								} else {
									value = pcxReader.ReadByte();
									if ((value & 0xC0) == 0xC0) {
										runCount = (byte)(value & 0x3F);
										runValue = pcxReader.ReadByte();
										value = runValue;
									}
								}
								break;
						}


						// Copy the byte to the decompressed scanline.
						decompressedScanline[x++] = value;

						// Have we reached the end of the scanline?
						if (x == decompressedScanline.Length) {
							x = 0;
							// Rearrange the pixel data in planes.
							if (planeCount > 1) {
								for (int rx = 0; rx < width; ++rx) {
									for (int plane = 0; plane < planeCount; ++plane) {
										rearrangedScanline[rx * planeCount + (planeCount - 1) - plane] = decompressedScanline[rx + width * plane];
									}
								}
							}
							// Copy the byte data.
							Marshal.Copy(rearrangedScanline, 0, outputPtr, width * planeCount);
							// Advance to next scanline in output image.
							outputPtr = new IntPtr(outputPtr.ToInt64() + lockedBitmap.Stride);
							++y;
						}
					}

				} finally {
					bitmap.UnlockBits(lockedBitmap);
				}

				// We have decompressed the image; convert to a PDN Document.
				return Document.FromImage(bitmap);
			}

			#endregion
		}

	}

}